/***************************************************************
   _ _         __           _   _                  ___   __
  (_) | ___   / / __   __ _| |_(_)_   _____ ___  _( _ ) / /_
  | | |/ _ \ / / '_ \ / _` | __| \ \ / / _ (_) \/ / _ \| '_ \
  | | | (_) / /| | | | (_| | |_| |\ V /  __/_ >  < (_) | (_) |
  |_|_|\___/_/ |_| |_|\__,_|\__|_| \_/ \___(_)_/\_\___/ \___/
  a tiny virtual computer : 64kw RAM, 32-bit, Dual Stack MISC
                    (c) charles childers
 **************************************************************/

#define V void
#define I int
#define C char
#define S short
#define L long
#define U unsigned

#include "font.h"
#include "image.h"

#define FW 320
#define FH 200
#define FONT_H 8
#define FONT_W 4

#define VGA_WIDTH 320
#define VGA_HEIGHT 240
#define VGA_MEMORY 0xA0000

V push(I);
I pop();


/* DRIVER: keyboard (in 386.s) ============================== */

I getchar();


/* DRIVER: interrupts ======================================= */

V term_putc(I);

V idt_flush(U I);

struct idt_entry_struct {
  U S base_lo;
  U S sel;
  U C always0;
  U C flags;
  U S base_hi;
} __attribute__((packed));
typedef struct idt_entry_struct idt_entry_t;

typedef struct registers {
  U I ds;
  U I edi, esi, ebp, esp, ebx, edx, ecx, eax;
  U I int_no, err_code;
  U I eip, cs, eflags, useresp, ss;
} registers_t;

struct idt_ptr_struct {
  U S limit;
  U I base;
} __attribute__((packed));
typedef struct idt_ptr_struct idt_ptr_t;

idt_entry_t idt_entries[256];
idt_ptr_t   idt_ptr;

V *memset(V *ptr, I value, I num) {
  unsigned char *p = ptr;
  while (num-- > 0) { *p++ = (U C)value; }
  return ptr;
}

V isr0();  V isr1();  V isr2();  V isr3();  V isr4();
V isr5();  V isr6();  V isr7();  V isr8();  V isr9();
V isr10(); V isr11(); V isr12(); V isr13(); V isr14();
V isr15(); V isr16(); V isr17(); V isr18(); V isr19();
V isr20(); V isr21(); V isr22(); V isr23(); V isr24();
V isr25(); V isr26(); V isr27(); V isr28(); V isr29();
V isr30(); V isr31();

V idt_set_gate(U C num, U I base, U S sel, U C flags) {
  idt_entries[num].base_lo = base & 0xFFFF;
  idt_entries[num].base_hi = (base >> 16) & 0xFFFF;
  idt_entries[num].sel     = sel;
  idt_entries[num].always0 = 0;
  idt_entries[num].flags   = flags;
}

V init_idt() {
  idt_ptr.limit = sizeof(idt_entry_t) * 256 -1;
  idt_ptr.base  = (U I)&idt_entries;

  memset(&idt_entries, 0, sizeof(idt_entry_t)*256);
  idt_set_gate( 0, (U I)isr0 , 0x08, 0x8E);
  idt_set_gate( 1, (U I)isr1 , 0x08, 0x8E);
  idt_set_gate( 2, (U I)isr2 , 0x08, 0x8E);
  idt_set_gate( 3, (U I)isr3 , 0x08, 0x8E);
  idt_set_gate( 4, (U I)isr4 , 0x08, 0x8E);
  idt_set_gate( 5, (U I)isr5 , 0x08, 0x8E);
  idt_set_gate( 6, (U I)isr6 , 0x08, 0x8E);
  idt_set_gate( 7, (U I)isr7 , 0x08, 0x8E);
  idt_set_gate( 8, (U I)isr8 , 0x08, 0x8E);
  idt_set_gate( 9, (U I)isr9 , 0x08, 0x8E);
  idt_set_gate(10, (U I)isr10, 0x08, 0x8E);
  idt_set_gate(11, (U I)isr11, 0x08, 0x8E);
  idt_set_gate(12, (U I)isr12, 0x08, 0x8E);
  idt_set_gate(13, (U I)isr13, 0x08, 0x8E);
  idt_set_gate(14, (U I)isr14, 0x08, 0x8E);
  idt_set_gate(15, (U I)isr15, 0x08, 0x8E);
  idt_set_gate(16, (U I)isr16, 0x08, 0x8E);
  idt_set_gate(17, (U I)isr17, 0x08, 0x8E);
  idt_set_gate(18, (U I)isr18, 0x08, 0x8E);
  idt_set_gate(19, (U I)isr19, 0x08, 0x8E);
  idt_set_gate(20, (U I)isr20, 0x08, 0x8E);
  idt_set_gate(21, (U I)isr21, 0x08, 0x8E);
  idt_set_gate(22, (U I)isr22, 0x08, 0x8E);
  idt_set_gate(23, (U I)isr23, 0x08, 0x8E);
  idt_set_gate(24, (U I)isr24, 0x08, 0x8E);
  idt_set_gate(25, (U I)isr25, 0x08, 0x8E);
  idt_set_gate(26, (U I)isr26, 0x08, 0x8E);
  idt_set_gate(27, (U I)isr27, 0x08, 0x8E);
  idt_set_gate(28, (U I)isr28, 0x08, 0x8E);
  idt_set_gate(29, (U I)isr29, 0x08, 0x8E);
  idt_set_gate(30, (U I)isr30, 0x08, 0x8E);
  idt_set_gate(31, (U I)isr31, 0x08, 0x8E);
  idt_flush((U I)&idt_ptr);
}

void isr_handler(registers_t regs) {
  if (regs.int_no == 0) {
    term_putc('\n');
    term_putc('D'); term_putc('B'); term_putc('Z');
    term_putc(':'); term_putc('H'); term_putc('A');
    term_putc('L'); term_putc('T');
    while (1) {}
  }
}


/* DRIVER: i/o ports ======================================== */

U C inb(U I port) {
  U C ret;
  asm volatile ("inb %%dx, %%al":"=a" (ret):"d" (port));
  return ret;
}

V outb(U I port, U C value) {
  asm volatile ("outb %%al, %%dx": :"d" (port), "a" (value));
}

static inline U S inw(U S port) {
  U S ret;
  asm volatile ( "inw %1, %0"
                 : "=a"(ret)
                 : "Nd"(port) );
  return ret;
}

static inline V outw(U S port, U S val)
{
  asm volatile ( "outw %0, %1" : : "a"(val), "Nd"(port) );
}


/* DRIVER: Physical RAM Interface =========================== */

V ram_store() {
  L address = pop();
  L value   = pop();
  *((I*)address) = value;
}

V ram_fetch() { push(*((I*)pop())); }

V ram_store_byte() {
  L address = pop();
  L value   = pop();
  *((C *)address) = (C)value;
}

V ram_fetch_byte() { push((L)*((C *)pop()) & 0xFF); }

V ram() {
  switch (pop()) {
    case 0: ram_fetch_byte(); break;
    case 1: ram_store_byte(); break;
    case 2: ram_fetch(); break;
    case 3: ram_store(); break;
    default: break;
  }
}


/* DRIVER: PIO-MODE (P)ATA ================================== */

#define ATA_READ           0x20
#define ATA_WRITE          0x30
#define ATA_FLUSH          0xE7
#define ATA_PRIMARY        0x1F0
#define ATA_DATA           0x1F0
#define ATA_ERROR          0x1F1
#define ATA_FEATURES       0x1F1
#define ATA_SECTOR_COUNT   0x1F2
#define ATA_SECTOR_NUMBER  0x1F3
#define ATA_CYLINDER_LOW   0x1F4
#define ATA_CYLINDER_HIGH  0x1F5
#define ATA_DRIVE          0x1F6
#define ATA_HEAD           0x1F6
#define ATA_STATUS         0x1F7
#define ATA_COMMAND        0x1F7

I sx, sa;

V ata_wait_bsy() { while (inb(ATA_STATUS) & 0x80); }
V ata_wait_drq() { while (!(inb(ATA_STATUS) & 0x40)); }

V ata_request(I n, U C req) {
  ata_wait_bsy();
  outb(ATA_HEAD,          0xE0 | ((n >> 24) &  0x0F));
  outb(ATA_ERROR,         0x00);
  outb(ATA_SECTOR_COUNT,  0x08);
  outb(ATA_SECTOR_NUMBER, (U C)n);
  outb(ATA_CYLINDER_LOW,  (U C)(n >> 8));
  outb(ATA_CYLINDER_HIGH, (U C)(n >> 16));
  outb(ATA_COMMAND, req);
}

V ata_read(I n, C *buffer) {
  ata_request(n, ATA_READ);
  for (sx = 0; sx < 8; sx++) {
    ata_wait_bsy();
    ata_wait_drq();
    for (sa = 0; sa < 512; sa += 2) {
      buffer[(sx * 512) + sa] = inw(ATA_PRIMARY);
    }
  }
}

V ata_write(I n, C *buffer) {
  ata_request(n, ATA_WRITE);
  for (sx = 0; sx < 8; sx++) {
    ata_wait_bsy();
    ata_wait_drq();
    for (sa = 0; sa < 512; sa += 2) {
      outw(ATA_PRIMARY, buffer[(sx * 512) + sa]);
    }
  }
}


/* DRIVER: VGA (320x200x256) ================================ */

#define VGA_AC_INDEX      0x3C0
#define VGA_AC_WRITE      0x3C0
#define VGA_AC_READ       0x3C1
#define VGA_INSTAT_READ   0x3DA
#define VGA_MISC_WRITE    0x3C2
#define VGA_MISC_READ     0x3CC

/*                        COLOR emulation  MONO emulation */
#define VGA_CRTC_INDEX    0x3D4            /* 0x3B4 */
#define VGA_CRTC_DATA     0x3D5            /* 0x3B5 */
#define VGA_GC_INDEX      0x3CE
#define VGA_GC_DATA       0x3CF
#define VGA_SEQ_INDEX     0x3C4
#define VGA_SEQ_DATA      0x3C5

#define VGA_NUM_AC_REGS   21
#define VGA_NUM_CRTC_REGS 25
#define VGA_NUM_GC_REGS   9
#define VGA_NUM_SEQ_REGS  5

U C g_320x200x256[] = {
/* MISC */  0x63,
/* SEQ  */  0x03, 0x01, 0x0F, 0x00, 0x0E,
/* CRTC */  0x5F, 0x4F, 0x50, 0x82, 0x54, 0x80, 0xBF, 0x1F,
            0x00, 0x41, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x9C, 0x0E, 0x8F, 0x28, 0x40, 0x96, 0xB9, 0xA3,
            0xFF,
/* GC */    0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0F,
            0xFF,
/* AC */    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
            0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
            0x41, 0x00, 0x0F, 0x00, 0x00
};

V putpixel(I x, I y, S color) {
  U S offset = x + FW * y;
  U C *VGA = (U C *)VGA_MEMORY;
  VGA[offset] = color ? 0x02 : 0x0;
}

I get_pixel(I x, I y) {
  U C *VGA = (U C *)VGA_MEMORY;
  if (x >= 0 && x < VGA_WIDTH && y >= 0 && y < VGA_HEIGHT) {
    I offset = y * VGA_WIDTH + x;
    return VGA[offset];
  }
  return 0;
}

V write_regs(U C *regs) {
  U i;

  /* write MISCELLANEOUS reg */
  outb(VGA_MISC_WRITE, *regs);
  regs++;

  /* write SEQUENCER regs */
  for(i = 0; i < VGA_NUM_SEQ_REGS; i++) {
    outb(VGA_SEQ_INDEX, i);
     outb(VGA_SEQ_DATA, *regs);
     regs++;
  }

  /* unlock CRTC registers */
  outb(VGA_CRTC_INDEX, 0x03);
  outb(VGA_CRTC_DATA, inb(VGA_CRTC_DATA) | 0x80);
  outb(VGA_CRTC_INDEX, 0x11);
  outb(VGA_CRTC_DATA, inb(VGA_CRTC_DATA) & ~0x80);

  /* make sure they remain unlocked */
  regs[0x03] |= 0x80;
  regs[0x11] &= ~0x80;

  /* write CRTC regs */
  for(i = 0; i < VGA_NUM_CRTC_REGS; i++) {
    outb(VGA_CRTC_INDEX, i);
    outb(VGA_CRTC_DATA, *regs);
    regs++;
  }

  /* write GRAPHICS CONTROLLER regs */
  for(i = 0; i < VGA_NUM_GC_REGS; i++) {
    outb(VGA_GC_INDEX, i);
    outb(VGA_GC_DATA, *regs);
    regs++;
  }

  /* write ATTRIBUTE CONTROLLER regs */
  for(i = 0; i < VGA_NUM_AC_REGS; i++) {
    inb(VGA_INSTAT_READ);
    outb(VGA_AC_INDEX, i);
    outb(VGA_AC_WRITE, *regs);
    regs++;
  }

  /* lock 16-color palette and unblank display */
  inb(VGA_INSTAT_READ);
  outb(VGA_AC_INDEX, 0x20);
}

V vga_clear() {
  for (I x = 0; x < FW; x++)
    for(I y = 0; y < FH; y++)
      putpixel(x, y, 0x0);
}

V vga_init() {
  write_regs(g_320x200x256);
  vga_clear();
}


/* DRIVER: Text Display ===================================== */

V dputc(I);

I acc[32], acc_i = 0, ground = 0, on = 0, tx, ty, ts;

V scroll() {
  I offset;
  U C *VGA = (U C *)VGA_MEMORY;
  for (I y = 8; y < VGA_HEIGHT; ++y) {
    for (I x = 0; x < VGA_WIDTH; ++x) {
      offset = y * VGA_WIDTH + x;
      I destinationOffset = (y - 8) * VGA_WIDTH + x;
      VGA[destinationOffset] = VGA[offset];
    }
  }
  for (I y = VGA_HEIGHT - 8; y < VGA_HEIGHT; ++y) {
    for (I x = 0; x < VGA_WIDTH; ++x) {
      VGA[y * VGA_WIDTH + x] = 0;
    }
  }
}

V cursor() {
  if (ty == 25) { scroll(); }
}

U C reverse(U C b) {
  b = (b & 0xF0) >> 4 | (b & 0x0F) << 4;
  b = (b & 0xCC) >> 2 | (b & 0x33) << 2;
  b = (b & 0xAA) >> 1 | (b & 0x55) << 1;
  return b;
}

V handle_special_character(I c) {
  if (c == 8 || c == 127) { /* BACKSPACE/DELETE */
    if (tx - FONT_W >= 0) {
      tx -= FONT_W ;
      dputc(' ');
      tx -= FONT_W;
    }
  } else if (c == 9) { /* TAB */
    I ts = tx / FONT_W;
    for (I i = ts; i < ((ts + 7) / 8) * 8; i++) { dputc(' '); }
  } else if (c == 10 || c == 13) { /* CR or LF */
    ty += FONT_H;
    tx = 0;
  }
}

V handle_regular_character(C *bitmap) {
  I x, y, set;
  for (x = 0; x < FONT_H; x++) {
    for (y = 0; y < FONT_W; y++) {
      set = reverse(bitmap[x]) & 1 << y;
      putpixel(tx + y, ty + x, set ? 0x02 : 0);
    }
  }
  tx += FONT_W;
}

V advance_cursor_and_scroll() {
  if (tx >= FW) { tx = 0; ty += FONT_H; }
  if (ty > (FH - FONT_H)) { ty -= FONT_H; scroll(); }
}

V dputc(I c) {
  C *bitmap = (C *)font + ((c + 15) * ((FONT_H * FONT_W)/4)) + 1030;

  handle_special_character(c);

  if (!(c == 8 || c == 127 || c == 9 || c == 10 || c == 13)) {
    handle_regular_character(bitmap);
  }

  advance_cursor_and_scroll();
  cursor();
}

I acc_pop() {
  I k = acc_i;
  if (acc_i > 0) {
    acc_i -= 1;
    if (acc_i == -1) {
      acc_i = 0;
    }
  }
  return acc[k];
}

V acc_reset() {
  for (I x = 0; x < 32; x++) { acc[x] = 0; }
  ground = on = 0;
}

V gt_up(I n)       { ty -= FONT_H; if (ty < 0) ty = 0;  }
V gt_down(I n)     { ty += FONT_H; if (ty > 24 * FONT_H) ty = 24 * FONT_H; }
V gt_left(I n)     { tx -= FONT_W; if (tx < 0) tx = 0; }
V gt_right(I n)    { tx += FONT_W; if (tx > 79 * FONT_W) ty = 79 * FONT_W; }
V gt_move_cursor() {
  if (acc_i >= 1) {
    I a = (acc_pop() - 1) * FONT_W;
    I b = (acc_pop() - 1) * FONT_H;
    ty = b; if (ty < 0) ty = 0; if (ty > FONT_H * 24) ty = FONT_H * 24;
    tx = a; if (tx < 0) tx = 0; if (tx > FONT_W * 79) tx = FONT_W * 79;
    acc_reset();
    return;
  }
  /* home */
  tx = 0; ty = 0;
  acc_reset();
}

V gt_reset() { acc_reset(); }
V gt_next()  { acc_i += 1; acc[acc_i] = 0; }
V gt_clear() { acc_reset(); vga_clear(); tx = ty = 0; }
V gt_set_attr() { acc_reset(); }

V term_putc(I c) {
  if (c == 27) {
    on = -2;
    return;
  }
  switch (on) {
    case 0:
      dputc(c);
      return;
    case -2:
      if (c == '[') {
        on = -1;
        return;
      }
      on = 0;
      dputc(c);
      return;
  }
  if (c >= '0' && c <= '9') {
    acc[acc_i] *= 10;
    acc[acc_i] += c - 48;
    if (c == '3') {
      ground = 0;
    }
    if (c == '4') {
      ground = 1;
    }
  } else {
    switch (c) {
      case 'A': gt_up(acc_pop());    break;
      case 'B': gt_down(acc_pop());  break;
      case 'C': gt_left(acc_pop());  break;
      case 'D': gt_right(acc_pop()); break;
      case 'm': gt_set_attr();       break;
      case ';': gt_next();           break;
      case 'H': gt_move_cursor();    break;
      case 'J': gt_clear();          break;
      default: on = 0;
               dputc(c);
    }
  }
}


/* ilo itself =============================================== */

#define T ds[sp]    /* Top of Data Stack    */
#define N ds[sp-1]  /* Next on Data Stack   */
#define R as[rp]    /* Top of Address Stack */

I ip,        /* instruction pointer             */
  sp,        /* data stack pointer              */
  rp,        /* address stack pointer           */
  ds[32],    /* data stack                      */
  as[256],   /* address stack                   */
  m[65536];  /* memory                          */

C *blocks,   /* name of block file (ilo.blocks) */
  *rom;      /* name of image (ilo.rom)         */

/* the other variables are used by the various
   functions for misc. purposes                   */

I a, b, f, s, d, l;
C i[1];

V push(I v) { ds[sp + 1] = v; sp += 1; }
I pop() { sp -= 1; return ds[sp + 1]; }

V load_image() {
  for (a = 0; a < 65536; a++) m[a] = image[a];
  ip = sp = rp = 0;
}

V save_image() {
}

V block_common() {
  b = pop(); /* block buffer */
  a = pop(); /* block number */
}

V read_block() {
  block_common();
  ata_read(a * 8, (C *)(m + b));
}

V write_block() {
  block_common();
  ata_write(a * 8, (C *)(m + b));
}

V save_ip() { rp += 1; R = ip; }
V symmetric() { if (b >= 0 && N < 0) { T += 1; N -= b; } }

V li() { ip += 1; push(m[ip]); }
V du() { push(T); }
V dr() { ds[sp] = 0; sp -= 1; }
V sw() { a = T; T = N; N = a; }
V pu() { rp += 1; R = pop(); }
V po() { push(R); rp -= 1; }
V ju() { ip = pop() - 1; }
V ca() { save_ip(); ip = pop() - 1; }
V cc() { a = pop(); if (pop()) { save_ip(); ip = a - 1; } }
V cj() { a = pop(); if (pop()) { ip = a - 1; } }
V re() { ip = R; rp -= 1; }
V eq() { N = (N == T) ? -1 : 0; sp -= 1; }
V ne() { N = (N != T) ? -1 : 0; sp -= 1; }
V lt() { N = (N <  T) ? -1 : 0; sp -= 1; }
V gt() { N = (N >  T) ? -1 : 0; sp -= 1; }
V fe() { T = m[T]; }
V st() { m[T] = N; sp -= 2; }
V ad() { N += T; sp -= 1; }
V su() { N -= T; sp -= 1; }
V mu() { N *= T; sp -= 1; }
V di() { a = T; b = N; T = b / a; N = b % a; symmetric(); }
V an() { N = T & N; sp -= 1; }
V or() { N = T | N; sp -= 1; }
V xo() { N = T ^ N; sp -= 1; }
V sl() { N = N << T; sp -= 1; }
V sr() { N = N >> T; sp -= 1; }
V cp() { l = pop(); d = pop(); s = T; T = -1;
         while (l) { if (m[d] != m[s]) { T = 0; }
                     l -= 1; s += 1; d += 1; } }
V cy() { l = pop(); d = pop(); s = pop();
         while (l) { m[d] = m[s]; l -= 1; s += 1; d += 1; } }
V ioa() { term_putc(pop()); }
V iob() { push(getchar()); term_putc(T); }
V ioc() { read_block(); }
V iod() { write_block(); }
V ioe() { save_image(); }
V iof() { load_image(); ip = -1; }
V iog() { ip = 65536; }
V ioh() { push(sp); push(rp); }
V pio() {
  switch (pop()) {
    case 0: a = pop(); push((I)inb(a)); break;
    case 1: a = pop(); push((I)inw((S)a)); break;
    case 2: a = pop(); b = pop(); outb(a, (C)b); break;
    case 3: a = pop(); b = pop(); outw(a, (S)b); break;
  }
}

V vga_pp() { I c = pop();  I y = pop();  I x = pop();
             putpixel(x, y, c); }
V vga_gp() { I y = pop(); I x = pop(); push(get_pixel(x, y)); }

V io() {
  switch (pop()) {
    case 0: ioa(); break;  case 1: iob(); break;
    case 2: ioc(); break;  case 3: iod(); break;
    case 4: ioe(); break;  case 5: iof(); break;
    case 6: iog(); break;  case 7: ioh(); break;
    case 33:  vga_pp(); break;  case 34:  vga_gp(); break;
    case 100: pio();    break;  case 101: ram();    break;
  }
}

V process(I o) {
  switch (o) {
    case  0:       break;   case  1: li(); break;
    case  2: du(); break;   case  3: dr(); break;
    case  4: sw(); break;   case  5: pu(); break;
    case  6: po(); break;   case  7: ju(); break;
    case  8: ca(); break;   case  9: cc(); break;
    case 10: cj(); break;   case 11: re(); break;
    case 12: eq(); break;   case 13: ne(); break;
    case 14: lt(); break;   case 15: gt(); break;
    case 16: fe(); break;   case 17: st(); break;
    case 18: ad(); break;   case 19: su(); break;
    case 20: mu(); break;   case 21: di(); break;
    case 22: an(); break;   case 23: or(); break;
    case 24: xo(); break;   case 25: sl(); break;
    case 26: sr(); break;   case 27: cp(); break;
    case 28: cy(); break;   case 29: io(); break;
    default: break;
  }
}

V process_bundle(I opcode) {
  process(opcode & 0xFF);
  process((opcode >> 8) & 0xFF);
  process((opcode >> 16) & 0xFF);
  process((opcode >> 24) & 0xFF);
}

V execute() {
  while (ip < 65536) {
    process_bundle(m[ip]);
    ip += 1;
  }
}

I main(I argc, C **argv) {
  vga_init();
  init_idt();
  load_image();
  execute();
#ifdef QEMU
  outw(0x604, 0x2000);
#endif
#ifdef VIRTUALBOX
  outw(0x4004, 0x3400);
#endif
  return 0;
}
