/***************************************************************
   _           _        _ _       _     _            _
  (_)_ __  ___| |_ __ _| | |     | |__ | | ___   ___| | _____
  | | '_ \/ __| __/ _` | | |_____| '_ \| |/ _ \ / __| |/ / __|
  | | | | \__ \ || (_| | | |_____| |_) | | (_) | (__|   <\__ \
  |_|_| |_|___/\__\__,_|_|_|     |_.__/|_|\___/ \___|_|\_\___/
   a tool for installing or updating the basic Konilo blocks

                 copyright (c) charles childers
 **************************************************************/

#define V void
#define I int
#define C char
#define S short
#define U unsigned

#include "blocks.h"

I getchar();

/* x86 i/o ports */

U C inb(U I port) {
  U C ret;
  asm volatile ("inb %%dx, %%al":"=a" (ret):"d" (port));
  return ret;
}

V outb(U I port, U C value) {
  asm volatile ("outb %%al, %%dx": :"d" (port), "a" (value));
}

static inline U S inw(U S port)
{
  U S ret;
  asm volatile ( "inw %1, %0"
                 : "=a"(ret)
                 : "Nd"(port) );
  return ret;
}

static inline V outw(U S port, U S val)
{
  asm volatile ( "outw %0, %1" : : "a"(val), "Nd"(port) );
}


/* (p)ata support */

#define ATA_READ           0x20
#define ATA_WRITE          0x30
#define ATA_FLUSH          0xE7
#define ATA_PRIMARY        0x1F0
#define ATA_DATA           0x1F0
#define ATA_ERROR          0x1F1
#define ATA_FEATURES       0x1F1
#define ATA_SECTOR_COUNT   0x1F2
#define ATA_SECTOR_NUMBER  0x1F3
#define ATA_CYLINDER_LOW   0x1F4
#define ATA_CYLINDER_HIGH  0x1F5
#define ATA_DRIVE          0x1F6
#define ATA_HEAD           0x1F6
#define ATA_STATUS         0x1F7
#define ATA_COMMAND        0x1F7

I sx, sa;

V ata_wait_bsy() { while (inb(ATA_STATUS) & 0x80); }
V ata_wait_drq() { while (!(inb(ATA_STATUS) & 0x40)); }

V ata_request(I n, U C req) {
  ata_wait_bsy();
  outb(ATA_HEAD,          0xE0 | ((n >> 24) &  0x0F));
  outb(ATA_ERROR,         0x00);
  outb(ATA_SECTOR_COUNT,  0x08);
  outb(ATA_SECTOR_NUMBER, (U C)n);
  outb(ATA_CYLINDER_LOW,  (U C)(n >> 8));
  outb(ATA_CYLINDER_HIGH, (U C)(n >> 16));
  outb(ATA_COMMAND, req);
}

V ata_read(I n, C *buffer) {
  ata_request(n, ATA_READ);
  for (sx = 0; sx < 8; sx++) {
    ata_wait_bsy();
    ata_wait_drq();
    for (sa = 0; sa < 512; sa += 2) {
      buffer[(sx * 512) + sa] = inw(ATA_PRIMARY);
    }
  }
}

V ata_write(I n, C *buffer) {
  ata_request(n, ATA_WRITE);
  for (sx = 0; sx < 8; sx++) {
    ata_wait_bsy();
    ata_wait_drq();
    for (sa = 0; sa < 512; sa += 2) {
      outw(ATA_PRIMARY, buffer[(sx * 512) + sa]);
    }
  }
}


/* basic vga text mode display support */

C* video  = (C*)0xb8000;

I tx, ty, ts;
I attr = 0x06;

V scroll() {
  for (ts = 0; ts < (80 * 24 * 2); ts += 2) {
    video[ts] = video[160 + ts];
  }
  for (; ts < (80 * 25 * 2); ts += 2) {
    video[ts] = 32;
  }

  for (ts = 1; ts < (80 * 24 * 2); ts += 2) {
    video[ts] = video[160 + ts];
  }
  for (; ts < (80 * 25 * 2); ts += 2) {
    video[ts] = attr;
  }
  ty = 24; tx = 0;
}

I cursor() {
  if (ty == 25) { scroll(); }

  U S pos = (ty * 80) + tx;
  outb(0x3D4, 0x0F);
  outb(0x3D5, (U S) (pos & 0xFF));
  outb(0x3D4, 0x0E);
  outb(0x3D5, (U S) ((pos >> 8) & 0xFF));

  return ((ty * 80) + tx) * 2;
}

V next() {
  tx += 1;
  if (tx == 80) { tx = 0; ty += 1; }
}

V display(I c) {
  if (c == 10 || c == 13) { tx = 0; ty += 1; }
  else { video[cursor()] = (C)c;
         video[cursor() + 1] = attr; next(); }
  cursor();
}

V vga_init() {
  attr = 0x06;
  for (tx = 0; tx < 80 * 25 * 2; tx += 2) { video[tx] = 32;   }
  for (tx = 1; tx < 80 * 25 * 2; tx += 2) { video[tx] = attr; }
  tx = 0; ty = 0;
  cursor();
}


/* limited dec terminal escape sequence support */

I acc[32], acc_i = 0, ground = 0, on = 0;

I acc_pop() {
  I k = acc_i;
  if (acc_i > 0) {
    acc_i -= 1;
    if (acc_i == -1) {
      acc_i = 0;
    }
  }
  return acc[k];
}

V acc_reset() {
  for (I x = 0; x < 32; x++) {
    acc[x] = 0;
  }
  ground = 0;
  on = 0;
}

V gt_up(I n)       { ty -= 1; if (ty < 0) ty = 0;   cursor(); }
V gt_down(I n)     { ty += 1; if (ty > 24) ty = 24; cursor(); }
V gt_left(I n)     { tx -= 1; if (tx < 0) tx = 0;   cursor(); }
V gt_right(I n)    { tx += 1; if (tx > 79) ty = 79; cursor(); }
V gt_move_cursor() {
  if (acc_i >= 1) {
    I a = acc_pop();
    I b = acc_pop();
    ty = b - 1; if (ty < 0) ty = 0; if (ty > 24) ty = 24;
    tx = a - 1; if (tx < 0) tx = 0; if (tx > 79) tx = 79;
    cursor();
    acc_reset();
    return;
  }
  /* home */
  acc_reset(); cursor();
}

V gt_set_fg(I c)   { c = c - 30; attr = attr & 0xF0;
                     attr = attr + c; acc_reset(); }
V gt_set_bg(I c)   { c = c - 40; attr = attr & 0x0F;
                     attr = attr + (c << 4); acc_reset(); }
V gt_reset()       { acc_reset(); attr = 0x06; }
V gt_next()        { acc_i += 1; acc[acc_i] = 0; }
V gt_clear()       { acc_reset(); vga_init(); }
V gt_set_attr()    {
  if (acc_pop() == 0) { gt_reset(); return; }
  ground == 0 ? gt_set_fg(acc_pop()) : gt_set_bg(acc_pop());
  acc_reset();
}

V term_putc(I c) {
  if (c == 27) {
    on = -2;
    return;
  }
  switch (on) {
    case 0:
      display(c);
      return;
    case -2:
      if (c == '[') {
        on = -1;
        return;
      }
      on = 0;
      display(c);
      return;
  }
  if (c >= '0' && c <= '9') {
    acc[acc_i] *= 10;
    acc[acc_i] += c - 48;
    if (c == '3') {
      ground = 0;
    }
    if (c == '4') {
      ground = 1;
    }
  } else {
    switch (c) {
      case 'A': gt_up(acc_pop());    break;
      case 'B': gt_down(acc_pop());  break;
      case 'C': gt_left(acc_pop());  break;
      case 'D': gt_right(acc_pop()); break;
      case 'm': gt_set_attr();       break;
      case ';': gt_next();           break;
      case 'H': gt_move_cursor();    break;
      case 'J': gt_clear();          break;
      default: on = 0;
               display(c);
    }
  }
  cursor();
}

V puts(C *s) {
  while (*s != '\0') {
    term_putc(*s);
    s++;
  }
}

I main(I argc, C **argv) {
  I k = ' ';
  I done = 0;
  I a = 0;
  C *b = (C*)image;

  vga_init();
  puts("(I)NSTALL or (U)PDATE BLOCKS or (A)BORT");
  while (done == 0) {
    k = getchar();
    if (k == 'I' || k == 'i' || k == 'U' || k == 'u') {
      done = 1;
    }
    if (k == 'A' || k == 'a') {
      done = 1;
    }
  }

  if (k == 'A' || k == 'a') {
    return 1;
  }

  if (k == 'I' || k == 'i') {
    puts("\nINSTALLING BLOCKS\n");
    a = 0;
  } else {
    puts("\nUPDATING BLOCKS\n");
    term_putc('.'); b += 4096;
    term_putc('.'); b += 4096;
    term_putc('.'); b += 4096;
    a = 3;
  }
  for(; a < 128; a++) {
    term_putc('+');
    for(I c = 0; c < 2000000000; c++);
    ata_write(a * 8, b);
    b += 4096;
  }
  term_putc('\n');
  puts("\nCOMPLETED. PRESS A KEY TO REBOOT.");
  getchar();
#ifdef QEMU
  outw(0x604, 0x2000);
#endif
#ifdef VIRTUALBOX
  outw(0x4004, 0x3400);
#endif
  return 0;
}
