/***************************************************************
             crc's            _ _      ___ _
              _ __ ___  _   _| | |_   / (_) | ___
             | '_ ` _ \| | | | | __| / /| | |/ _ \
             | | | | | | |_| | | |_ / / | | | (_) |
             |_| |_| |_|\__,_|_|\__/_/  |_|_|\___/
                        for native x86

                     (c) charles childers

 This is a special implementation of my ilo computer, with many
 copies of ilo running under a single process. Under this you
 can quickly switch between ilo instances.

 - F9 restarts the current instance
 - F1 to F5 to switch to a different instance
 - F11 switch to previous instance
 - F12 switch to next instance
 **************************************************************/

#include "image.h"
#include "font.h"

#define V void
#define I int
#define C char
#define UI unsigned int
#define US unsigned short
#define UC unsigned char
#define U unsigned

#define ILOS 5

typedef struct ilo ilo;

struct ilo {
  I  ip,        /* instruction pointer             */
     sp,        /* data stack pointer              */
     rp,        /* address stack pointer           */
     ds[33],    /* data stack                      */
     as[257],   /* address stack                   */
     m[65536];  /* memory                          */
  C *blocks,    /* name of block file (ilo.blocks) */
    *rom;       /* name of image (ilo.rom)         */
  I  a, b, f, s, d, l;
  C  i[1];
};

struct ilo systems[ILOS];

I cur, cancel;

#define DS v->ds
#define AS v->as
#define SP v->sp
#define RP v->rp
#define IP v->ip
#define M  v->m
#define A  v->a
#define B  v->b
#define S  v->s
#define D  v->d
#define L  v->l
#define T DS[SP]    /* Top of Data Stack    */
#define N DS[SP-1]  /* Next on Data Stack   */
#define R AS[RP]    /* Top of Address Stack */

#define FW 320
#define FH 200
#define FONT_H 8
#define FONT_W 4

#define VGA_WIDTH 320
#define VGA_HEIGHT 240
#define VGA_MEMORY 0xA0000

void push(ilo*, int);
int pop(ilo*);


/* DRIVER: Keyboard (in 386.s) ============================== */

I getchar();


/* DRIVER: interrupts ======================================= */

V idt_flush(UI);

struct idt_entry_struct {
  US base_lo;
  US sel;
  UC always0;
  UC flags;
  US base_hi;
} __attribute__((packed));
typedef struct idt_entry_struct idt_entry_t;

typedef struct registers {
  UI ds;
  UI edi, esi, ebp, esp, ebx, edx, ecx, eax;
  UI int_no, err_code;
  UI eip, cs, eflags, useresp, ss;
} registers_t;

struct idt_ptr_struct {
  US limit;
  UI base;
} __attribute__((packed));
typedef struct idt_ptr_struct idt_ptr_t;

idt_entry_t idt_entries[256];
idt_ptr_t   idt_ptr;

V *memset(V *ptr, I value, I num) {
  unsigned char *p = ptr;
  while (num-- > 0) { *p++ = (UC)value; }
  return ptr;
}

V isr0();  V isr1();  V isr2();  V isr3();  V isr4();
V isr5();  V isr6();  V isr7();  V isr8();  V isr9();
V isr10(); V isr11(); V isr12(); V isr13(); V isr14();
V isr15(); V isr16(); V isr17(); V isr18(); V isr19();
V isr20(); V isr21(); V isr22(); V isr23(); V isr24();
V isr25(); V isr26(); V isr27(); V isr28(); V isr29();
V isr30(); V isr31();

V idt_set_gate(UC num, UI base, US sel, UC flags) {
  idt_entries[num].base_lo = base & 0xFFFF;
  idt_entries[num].base_hi = (base >> 16) & 0xFFFF;
  idt_entries[num].sel     = sel;
  idt_entries[num].always0 = 0;
  idt_entries[num].flags   = flags;
}

V init_idt() {
  idt_ptr.limit = sizeof(idt_entry_t) * 256 -1;
  idt_ptr.base  = (UI)&idt_entries;

  memset(&idt_entries, 0, sizeof(idt_entry_t)*256);
  idt_set_gate( 0, (UI)isr0 , 0x08, 0x8E);
  idt_set_gate( 1, (UI)isr1 , 0x08, 0x8E);
  idt_set_gate( 2, (UI)isr2 , 0x08, 0x8E);
  idt_set_gate( 3, (UI)isr3 , 0x08, 0x8E);
  idt_set_gate( 4, (UI)isr4 , 0x08, 0x8E);
  idt_set_gate( 5, (UI)isr5 , 0x08, 0x8E);
  idt_set_gate( 6, (UI)isr6 , 0x08, 0x8E);
  idt_set_gate( 7, (UI)isr7 , 0x08, 0x8E);
  idt_set_gate( 8, (UI)isr8 , 0x08, 0x8E);
  idt_set_gate( 9, (UI)isr9 , 0x08, 0x8E);
  idt_set_gate(10, (UI)isr10, 0x08, 0x8E);
  idt_set_gate(11, (UI)isr11, 0x08, 0x8E);
  idt_set_gate(12, (UI)isr12, 0x08, 0x8E);
  idt_set_gate(13, (UI)isr13, 0x08, 0x8E);
  idt_set_gate(14, (UI)isr14, 0x08, 0x8E);
  idt_set_gate(15, (UI)isr15, 0x08, 0x8E);
  idt_set_gate(16, (UI)isr16, 0x08, 0x8E);
  idt_set_gate(17, (UI)isr17, 0x08, 0x8E);
  idt_set_gate(18, (UI)isr18, 0x08, 0x8E);
  idt_set_gate(19, (UI)isr19, 0x08, 0x8E);
  idt_set_gate(20, (UI)isr20, 0x08, 0x8E);
  idt_set_gate(21, (UI)isr21, 0x08, 0x8E);
  idt_set_gate(22, (UI)isr22, 0x08, 0x8E);
  idt_set_gate(23, (UI)isr23, 0x08, 0x8E);
  idt_set_gate(24, (UI)isr24, 0x08, 0x8E);
  idt_set_gate(25, (UI)isr25, 0x08, 0x8E);
  idt_set_gate(26, (UI)isr26, 0x08, 0x8E);
  idt_set_gate(27, (UI)isr27, 0x08, 0x8E);
  idt_set_gate(28, (UI)isr28, 0x08, 0x8E);
  idt_set_gate(29, (UI)isr29, 0x08, 0x8E);
  idt_set_gate(30, (UI)isr30, 0x08, 0x8E);
  idt_set_gate(31, (UI)isr31, 0x08, 0x8E);
  idt_flush((UI)&idt_ptr);
}

void isr_handler(registers_t regs) {
  if (regs.int_no == 0) { /* DBZ */
    while (1) {}
  }
}


/* DRIVER: i/o ports ======================================== */

unsigned char inb(unsigned int port) {
  unsigned char ret;
  asm volatile ("inb %%dx, %%al":"=a" (ret):"d" (port));
  return ret;
}

void outb(unsigned int port, unsigned char value) {
  asm volatile ("outb %%al, %%dx": :"d" (port), "a" (value));
}

static inline unsigned short inw(unsigned short port) {
  unsigned short ret;
  asm volatile ( "inw %1, %0"
                 : "=a"(ret)
                 : "Nd"(port) );
  return ret;
}

static inline void outw(unsigned short port, unsigned short val)
{
  asm volatile ( "outw %0, %1" : : "a"(val), "Nd"(port) );
}


/* DRIVER: PIO (P)ATA ======================================= */

#define ATA_READ           0x20
#define ATA_WRITE          0x30
#define ATA_FLUSH          0xE7
#define ATA_PRIMARY        0x1F0
#define ATA_DATA           0x1F0
#define ATA_ERROR          0x1F1
#define ATA_FEATURES       0x1F1
#define ATA_SECTOR_COUNT   0x1F2
#define ATA_SECTOR_NUMBER  0x1F3
#define ATA_CYLINDER_LOW   0x1F4
#define ATA_CYLINDER_HIGH  0x1F5
#define ATA_DRIVE          0x1F6
#define ATA_HEAD           0x1F6
#define ATA_STATUS         0x1F7
#define ATA_COMMAND        0x1F7

int sx, sa;

void ata_wait_bsy() { while (inb(ATA_STATUS) & 0x80); }
void ata_wait_drq() { while (!(inb(ATA_STATUS) & 0x40)); }

void ata_request(int n, unsigned char req) {
  ata_wait_bsy();
  outb(ATA_HEAD,          0xE0 | ((n >> 24) &  0x0F));
  outb(ATA_ERROR,         0x00);
  outb(ATA_SECTOR_COUNT,  0x08);
  outb(ATA_SECTOR_NUMBER, (unsigned char)n);
  outb(ATA_CYLINDER_LOW,  (unsigned char)(n >> 8));
  outb(ATA_CYLINDER_HIGH, (unsigned char)(n >> 16));
  outb(ATA_COMMAND, req);
}

void ata_read(int n, char *buffer) {
  ata_request(n, ATA_READ);
  for (sx = 0; sx < 8; sx++) {
    ata_wait_bsy();
    ata_wait_drq();
    for (sa = 0; sa < 512; sa += 2) {
      buffer[(sx * 512) + sa] = inw(ATA_PRIMARY);
    }
  }
}

void ata_write(int n, char *buffer) {
  ata_request(n, ATA_WRITE);
  for (sx = 0; sx < 8; sx++) {
    ata_wait_bsy();
    ata_wait_drq();
    for (sa = 0; sa < 512; sa += 2) {
      outw(ATA_PRIMARY, buffer[(sx * 512) + sa]);
    }
  }
}


/* DRIVER: Physical RAM ===================================== */

void ram_store(ilo *v) {
  long address = pop(v);
  long value   = pop(v);
  *((int*)address) = value;
}

void ram_fetch(ilo *v) { push(v, *((int*)pop(v))); }

void ram_store_byte(ilo *v) {
  long address = pop(v);
  long value   = pop(v);
  *((char *)address) = (char)value;
}

void ram_fetch_byte(ilo *v) {
  push(v, (long)*((char*)pop(v)) & 0xFF);
}

void ram(ilo *v) {
  switch (pop(v)) {
    case 0: ram_fetch_byte(v); break;
    case 1: ram_store_byte(v); break;
    case 2: ram_fetch(v); break;
    case 3: ram_store(v); break;
    default: break;
  }
}


/* DRIVER: VGA (320x200x256) ================================ */

#define VGA_AC_INDEX      0x3C0
#define VGA_AC_WRITE      0x3C0
#define VGA_AC_READ       0x3C1
#define VGA_INSTAT_READ   0x3DA
#define VGA_MISC_WRITE    0x3C2
#define VGA_MISC_READ     0x3CC

/*                        COLOR emulation  MONO emulation */
#define VGA_CRTC_INDEX    0x3D4            /* 0x3B4 */
#define VGA_CRTC_DATA     0x3D5            /* 0x3B5 */
#define VGA_GC_INDEX      0x3CE
#define VGA_GC_DATA       0x3CF
#define VGA_SEQ_INDEX     0x3C4
#define VGA_SEQ_DATA      0x3C5

#define VGA_NUM_AC_REGS   21
#define VGA_NUM_CRTC_REGS 25
#define VGA_NUM_GC_REGS   9
#define VGA_NUM_SEQ_REGS  5

U C g_320x200x256[] = {
/* MISC */  0x63,
/* SEQ  */  0x03, 0x01, 0x0F, 0x00, 0x0E,
/* CRTC */  0x5F, 0x4F, 0x50, 0x82, 0x54, 0x80, 0xBF, 0x1F,
            0x00, 0x41, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x9C, 0x0E, 0x8F, 0x28, 0x40, 0x96, 0xB9, 0xA3,
            0xFF,
/* GC */    0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0F,
            0xFF,
/* AC */    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
            0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
            0x41, 0x00, 0x0F, 0x00, 0x00
};

V putpixel(I x, I y, short color) {
  US offset = x + FW * y;
  UC *VGA = (UC *)VGA_MEMORY;
  VGA[offset] = color ? 0x02 : 0x0;
}

I get_pixel(I x, I y) {
  U C *VGA = (U C *)VGA_MEMORY;
  if (x >= 0 && x < VGA_WIDTH && y >= 0 && y < VGA_HEIGHT) {
    I offset = y * VGA_WIDTH + x;
    return VGA[offset];
  }
  return 0;
}

V write_regs(U C *regs) {
  U i;

  /* write MISCELLANEOUS reg */
  outb(VGA_MISC_WRITE, *regs);
  regs++;

  /* write SEQUENCER regs */
  for(i = 0; i < VGA_NUM_SEQ_REGS; i++) {
    outb(VGA_SEQ_INDEX, i);
     outb(VGA_SEQ_DATA, *regs);
     regs++;
  }

  /* unlock CRTC registers */
  outb(VGA_CRTC_INDEX, 0x03);
  outb(VGA_CRTC_DATA, inb(VGA_CRTC_DATA) | 0x80);
  outb(VGA_CRTC_INDEX, 0x11);
  outb(VGA_CRTC_DATA, inb(VGA_CRTC_DATA) & ~0x80);

  /* make sure they remain unlocked */
  regs[0x03] |= 0x80;
  regs[0x11] &= ~0x80;

  /* write CRTC regs */
  for(i = 0; i < VGA_NUM_CRTC_REGS; i++) {
    outb(VGA_CRTC_INDEX, i);
    outb(VGA_CRTC_DATA, *regs);
    regs++;
  }

  /* write GRAPHICS CONTROLLER regs */
  for(i = 0; i < VGA_NUM_GC_REGS; i++) {
    outb(VGA_GC_INDEX, i);
    outb(VGA_GC_DATA, *regs);
    regs++;
  }

  /* write ATTRIBUTE CONTROLLER regs */
  for(i = 0; i < VGA_NUM_AC_REGS; i++) {
    inb(VGA_INSTAT_READ);
    outb(VGA_AC_INDEX, i);
    outb(VGA_AC_WRITE, *regs);
    regs++;
  }

  /* lock 16-color palette and unblank display */
  inb(VGA_INSTAT_READ);
  outb(VGA_AC_INDEX, 0x20);
}

V vga_clear() {
  for (I x = 0; x < FW; x++)
    for(I y = 0; y < FH; y++)
      putpixel(x, y, 0x0);
}

V vga_init() {
  write_regs(g_320x200x256);
  vga_clear();
}


/* DRIVER: Text Display ===================================== */

V dputc(I);

I acc[32], acc_i = 0, ground = 0, on = 0, tx, ty, ts;

V scroll() {
  I offset;
  U C *VGA = (U C *)VGA_MEMORY;
  for (I y = 8; y < VGA_HEIGHT; ++y) {
    for (I x = 0; x < VGA_WIDTH; ++x) {
      offset = y * VGA_WIDTH + x;
      I destinationOffset = (y - 8) * VGA_WIDTH + x;
      VGA[destinationOffset] = VGA[offset];
    }
  }
  for (I y = VGA_HEIGHT - 8; y < VGA_HEIGHT; ++y) {
    for (I x = 0; x < VGA_WIDTH; ++x) {
      VGA[y * VGA_WIDTH + x] = 0;
    }
  }
}

V cursor() {
  if (ty == 25) { scroll(); }
}

U C reverse(U C b) {
  b = (b & 0xF0) >> 4 | (b & 0x0F) << 4;
  b = (b & 0xCC) >> 2 | (b & 0x33) << 2;
  b = (b & 0xAA) >> 1 | (b & 0x55) << 1;
  return b;
}

V handle_special_character(I c) {
  if (c == 8 || c == 127) { /* BACKSPACE/DELETE */
    if (tx - FONT_W >= 0) {
      tx -= FONT_W ;
      dputc(' ');
      tx -= FONT_W;
    }
  } else if (c == 9) { /* TAB */
    I ts = tx / FONT_W;
    for (I i = ts; i < ((ts + 7) / 8) * 8; i++) { dputc(' '); }
  } else if (c == 10 || c == 13) { /* CR or LF */
    ty += FONT_H;
    tx = 0;
  }
}

V handle_regular_character(C *bitmap) {
  I x, y, set;
  for (x = 0; x < FONT_H; x++) {
    for (y = 0; y < FONT_W; y++) {
      set = reverse(bitmap[x]) & 1 << y;
      putpixel(tx + y, ty + x, set ? 0x02 : 0);
    }
  }
  tx += FONT_W;
}

V advance_cursor_and_scroll() {
  if (tx >= FW) { tx = 0; ty += FONT_H; }
  if (ty > (FH - FONT_H)) { ty -= FONT_H; scroll(); }
}

V dputc(I c) {
  C *bitmap = (C *)font + ((c + 15) * ((FONT_H * FONT_W)/4)) + 1030;

  handle_special_character(c);

  if (!(c == 8 || c == 127 || c == 9 || c == 10 || c == 13)) {
    handle_regular_character(bitmap);
  }

  advance_cursor_and_scroll();
  cursor();
}

I acc_pop() {
  I k = acc_i;
  if (acc_i > 0) {
    acc_i -= 1;
    if (acc_i == -1) {
      acc_i = 0;
    }
  }
  return acc[k];
}

V acc_reset() {
  for (I x = 0; x < 32; x++) { acc[x] = 0; }
  ground = on = 0;
}

V gt_up(I n)       { ty -= FONT_H; if (ty < 0) ty = 0;  }
V gt_down(I n)     { ty += FONT_H; if (ty > 24 * FONT_H) ty = 24 * FONT_H; }
V gt_left(I n)     { tx -= FONT_W; if (tx < 0) tx = 0; }
V gt_right(I n)    { tx += FONT_W; if (tx > 79 * FONT_W) ty = 79 * FONT_W; }
V gt_move_cursor() {
  if (acc_i >= 1) {
    I a = (acc_pop() - 1) * FONT_W;
    I b = (acc_pop() - 1) * FONT_H;
    ty = b; if (ty < 0) ty = 0; if (ty > FONT_H * 24) ty = FONT_H * 24;
    tx = a; if (tx < 0) tx = 0; if (tx > FONT_W * 79) tx = FONT_W * 79;
    acc_reset();
    return;
  }
  /* home */
  tx = 0; ty = 0;
  acc_reset();
}

V gt_reset() { acc_reset(); }
V gt_next()  { acc_i += 1; acc[acc_i] = 0; }
V gt_clear() { acc_reset(); vga_clear(); tx = ty = 0; }
V gt_set_attr() { acc_reset(); }

V term_putc(ilo *v, I c) {
  if (c == 27) {
    on = -2;
    return;
  }
  switch (on) {
    case 0:
      dputc(c);
      return;
    case -2:
      if (c == '[') {
        on = -1;
        return;
      }
      on = 0;
      dputc(c);
      return;
  }
  if (c >= '0' && c <= '9') {
    acc[acc_i] *= 10;
    acc[acc_i] += c - 48;
    if (c == '3') {
      ground = 0;
    }
    if (c == '4') {
      ground = 1;
    }
  } else {
    switch (c) {
      case 'A': gt_up(acc_pop());    break;
      case 'B': gt_down(acc_pop());  break;
      case 'C': gt_left(acc_pop());  break;
      case 'D': gt_right(acc_pop()); break;
      case 'm': gt_set_attr();       break;
      case ';': gt_next();           break;
      case 'H': gt_move_cursor();    break;
      case 'J': gt_clear();          break;
      default: on = 0;
               dputc(c);
    }
  }
}


/* ilo ====================================================== */

V *memcpy(void *dst, const void *src, unsigned long n) {
  void *ret = dst;
  asm volatile("rep movsb" : "+D" (dst)
                           : "c"(n), "S"(src) : "cc", "memory");
  return ret;
}

V push(ilo *v, I n) { DS[SP + 1] = n; SP += 1; }
I pop(ilo *v) { SP -= 1; return DS[SP + 1]; }

V load_image(ilo *v) {
  for (A = 0; A < 65536; A++)
    M[A] = image[A];
  IP = SP = RP = 0;
}

V save_image(ilo *v) { }

V block_common(ilo *v) {
  B = pop(v); /* block buffer */
  A = pop(v); /* block number */
}

V read_block(ilo *v) {
  block_common(v);
  ata_read(A * 8, (char *)(M + B));
}

V write_block(ilo *v) {
  block_common(v);
  ata_write(A * 8, (char *)(M + B));
}

V save_ip(ilo *v) { RP += 1; R = IP; }
V symmetric(ilo *v) { if (B >= 0 && N < 0) { T += 1; N -= B; } }

V li(ilo *v) { IP += 1; push(v, M[IP]); }
V du(ilo *v) { push(v, T); }
V dr(ilo *v) { DS[SP] = 0; SP -= 1; }
V sw(ilo *v) { A = T; T = N; N = A; }
V pu(ilo *v) { RP += 1; R = pop(v); }
V po(ilo *v) { push(v, R); RP -= 1; }
V ju(ilo *v) { IP = pop(v) - 1; }
V ca(ilo *v) { save_ip(v); IP = pop(v) - 1; }
V cc(ilo *v) { A = pop(v);
                if (pop(v)) { save_ip(v);
                               IP = A - 1; } }
V cj(ilo *v) { A = pop(v);
                if (pop(v)) { IP = A - 1; } }
V re(ilo *v) { IP = R; RP -= 1; }
V eq(ilo *v) { N = (N == T) ? -1 : 0; SP -= 1; }
V ne(ilo *v) { N = (N != T) ? -1 : 0; SP -= 1; }
V lt(ilo *v) { N = (N <  T) ? -1 : 0; SP -= 1; }
V gt(ilo *v) { N = (N >  T) ? -1 : 0; SP -= 1; }
V fe(ilo *v) { T = M[T]; }
V st(ilo *v) { M[T] = N; SP -= 2; }
V ad(ilo *v) { N += T; SP -= 1; }
V su(ilo *v) { N -= T; SP -= 1; }
V mu(ilo *v) { N *= T; SP -= 1; }
V di(ilo *v) { A = T; B = N;
               T = B / A; N = B % A;
               symmetric(v); }
V an(ilo *v) { N = T & N; SP -= 1; }
V or(ilo *v) { N = T | N; SP -= 1; }
V xo(ilo *v) { N = T ^ N; SP -= 1; }
V sl(ilo *v) { N = N << T; SP -= 1; }
V sr(ilo *v) { N = N >> T; SP -= 1; }
V cp(ilo *v) { L = pop(v); D = pop(v); S = T; T = -1;
  while (L) { if (M[D] != M[S]) { T = 0; }
                  L -= 1; S += 1; D += 1; } }
V cy(ilo *v) { L = pop(v); D = pop(v); S = pop(v);
  while (L) { M[D] = M[S];
                 L -= 1; S += 1; D += 1; } }

V ioa(ilo *v) { term_putc(v, pop(v)); }
V iob(ilo *v) {
  I a = getchar();
  switch (a) {
    case -1: cur = 0; a = 32; break;
    case -2: cur = 1; a = 32; break;
    case -3: cur = 2; a = 32; break;
    case -4: cur = 3; a = 32; break;
    case -5: cur = 4; a = 32; break;
    case -9: load_image(v); IP = -1; break;
    case -29: cur -= 1; a = 32; if (cur < 0) cur = 4; break;
    case -30: cur += 1; a = 32; if (cur > 4) cur = 0; break;
  }
  push(v, a); term_putc(v, T);
}
V ioc(ilo *v) { read_block(v); }
V iod(ilo *v) { write_block(v); }
V ioe(ilo *v) { save_image(v); }
V iof(ilo *v) { load_image(v); IP = -1; }
V iog(ilo *v) { IP = 65536; }
V ioh(ilo *v) { push(v, SP); push(v, RP); }
V iox(ilo *v) {
  switch (pop(v)) {
    case 0: A = pop(v); push(v, (int)inb(A)); break;
    case 1: A = pop(v); push(v, (int)inw((short)A)); break;
    case 2: A = pop(v); B = pop(v); outb(A, (C)B); break;
    case 3: A = pop(v); B = pop(v); outw(A, (short)B); break;
  }
}
V io(ilo *v) {
  switch (pop(v)) {
    case 0: ioa(v); break;  case 1: iob(v); break;
    case 2: ioc(v); break;  case 3: iod(v); break;
    case 4: ioe(v); break;  case 5: iof(v); break;
    case 6: iog(v); break;  case 7: ioh(v); break;
    case 100: iox(v); break;
    case 101: ram(v); break;
  }
}

V process(ilo *v, I o) {
  switch (o) {
    case  0:       break;   case  1: li(v); break;
    case  2: du(v); break;   case  3: dr(v); break;
    case  4: sw(v); break;   case  5: pu(v); break;
    case  6: po(v); break;   case  7: ju(v); break;
    case  8: ca(v); break;   case  9: cc(v); break;
    case 10: cj(v); break;   case 11: re(v); break;
    case 12: eq(v); break;   case 13: ne(v); break;
    case 14: lt(v); break;   case 15: gt(v); break;
    case 16: fe(v); break;   case 17: st(v); break;
    case 18: ad(v); break;   case 19: su(v); break;
    case 20: mu(v); break;   case 21: di(v); break;
    case 22: an(v); break;   case 23: or(v); break;
    case 24: xo(v); break;   case 25: sl(v); break;
    case 26: sr(v); break;   case 27: cp(v); break;
    case 28: cy(v); break;   case 29: io(v); break;
    default: break;
  }
}

V process_bundle(ilo *v, I opcode) {
  process(v, opcode & 0xFF);
  process(v, (opcode >> 8) & 0xFF);
  process(v, (opcode >> 16) & 0xFF);
  process(v, (opcode >> 24) & 0xFF);
}

V execute() {
  ilo *v = &systems[cur];
  while (IP < 65536) {
    cancel = 0;
    process_bundle(v, M[IP]);
    IP += 1;
    v = &systems[cur];
  }
}

I main(I argc, C **argv) {
  ilo *v;
  I i;
  cur = 0;
  init_idt();
  vga_init();
  for (i = 0; i < ILOS; i++) {
    v = &systems[i];
    load_image(v);
    term_putc(v, '0' + i);
  }
  term_putc(v, '\n');
  execute();
#ifdef QEMU
  outw(0x604, 0x2000);
#endif
#ifdef VIRTUALBOX
  outw(0x4004, 0x3400);
#endif
  return 0;
}
